#include <cmath>
#include <vector>
#include <cstdio>

#include <GL/gl.h>

#include "landform.h"
#include "terrain.h"

////////////////
//  Landform  //
////////////////

Landform::Landform(){
    this->rows = 0;
    this->cols = 0;
    this->width = 0.0;
    this->depth = 0.0;
    
    this->height = 0.0;
}

Landform::~Landform(){

}

Point Landform::getMeshDomainPoint(int i, int j){
    float w_inc = width/rows;
    float d_inc = depth/cols;

    return Point( (-width/2) + (i*w_inc), 0.11, (-depth/2) + (j*d_inc) );
}

void Landform::projectBaseToMesh(){
    int row, col;
    double w_inc = width / rows;
    double d_inc = depth / cols;

    for (int i = 0; i < base.size(); i++){
        row = int( 0.100 + ( base[i].getX() + (width / 2.0) ) / w_inc);
        col = int( 0.100 + ( base[i].getZ() + (depth / 2.0) ) / d_inc);
        meshBase.addPoint( getMeshDomainPoint(row, col) );
    }
}

void Landform::projectMaxisToMesh(){
    int row, col;
    double w_inc = width / rows;
    double d_inc = depth / cols;

    for (int i = 0; i < maxis.size(); i++){
        row = int( 0.100 + ( maxis[i].getX() + (width / 2.0) ) / w_inc);
        col = int( 0.100 + ( maxis[i].getZ() + (depth / 2.0) ) / d_inc);
        meshMaxis.addPoint( getMeshDomainPoint(row, col) );
    }
    maxis = meshMaxis;
}

Trace Landform::removeRedundancy(Trace trace){
    Trace projBase = trace;
    trace.clear();
    
    trace.addPoint( projBase[-1] );
    for (int i = 0; i < projBase.size(); i++){
        if ( (trace[-1]  - projBase[i]).lenght() > 0.1 )
            trace.addPoint(projBase[i]);
    }
    return trace;
}

void Landform::completeBaseInMesh(Trace trace){
    double w_inc = width / rows;
    double d_inc = depth / cols;
    int r_inc;
    float z, z_t; 
    float tj;
    int  c_inc;
    float x, x_t;
    
    adjusted.addPoint(trace[0]);
    for (int i = 1; i < trace.size(); i++){
        r_inc = int (0.100 + ( fabs(trace[i-1].getX() - trace[i].getX()) / w_inc ) );
        if ( r_inc > 1 ){
            for (int j = 1; j <= r_inc; j++){
                tj = float(j) / float(r_inc);
                z = ( (1 - tj) * trace[i-1].getZ() ) + ( tj * trace[i].getZ() );
                
                if ( trace[i-1].getX() < trace[i].getX() ){
                    x_t = adjusted[-1].getX() + w_inc;
                }
                else{
                    x_t = adjusted[-1].getX() - w_inc;
                }
                adjusted.addPoint( Point( x_t, trace[i-1].getY(), 
                                         adjusted[-1].getZ() )  );
                if ( fabs (z - adjusted[-1].getZ() ) > d_inc/2.0 ){
                    if ( trace[i-1].getZ() < trace[i].getZ() ){
                        z_t = adjusted[-1].getZ() + d_inc;
                    }
                    else{
                        z_t = adjusted[-1].getZ() - d_inc;
                    }
                    adjusted.addPoint( Point( adjusted[-1].getX(), 
                                             trace[i-1].getY(), z_t )  );
                }
            }
        }
        else{
            c_inc = int (0.100 + ( fabs(trace[i-1].getZ() - trace[i].getZ()) / d_inc ) );
            if ( c_inc >= 1 ){
                for (int j = 1; j <= c_inc; j++){
                    tj = float(j) / float(c_inc);
                    x = (1 - tj)* trace[i-1].getX() + tj * trace[i].getX();
                    
                    if ( trace[i-1].getZ() < trace[i].getZ() ){
                        z_t = adjusted[-1].getZ() + d_inc;
                    }
                    else{
                        z_t = adjusted[-1].getZ() - d_inc;
                    }
                    adjusted.addPoint( Point( adjusted[-1].getX(), 
                                             trace[i-1].getY(), z_t )  );
                    if ( fabs (x - adjusted[-1].getX() ) > w_inc/2.0 ){
                        if ( trace[i-1].getX() < trace[i].getX() ){
                            x_t = adjusted[-1].getX() + w_inc;
                        }
                        else{
                            x_t = adjusted[-1].getX() - w_inc;
                        }
                        adjusted.addPoint( Point( x_t, trace[i-1].getY(), 
                                                 adjusted[-1].getZ() )  );
                    }
                }
            }
            else{
                adjusted.addPoint( trace[i] );
            }
        }
    }
}

Point Landform::baseProjection(Point point, Point normal){
    Point p;
    Point proj;
    float dist = 10000;
    int index = -1;
    
    vector<Point> maxisPoint = base.getSkeleton().getGeometricPoints();
    for (int i = 0; i < maxisPoint.size(); i++){
        p = maxisPoint[i] - point;
        proj = normal * ( p.dotProduct(normal)  /  normal.dotProduct(normal) );
        
        if ( proj.distance(p) < 0.2 and proj.distance(point) < dist
        and proj.getX() * normal.getX() >= 0.0 and proj.getZ() * normal.getZ() >= 0.0 ){
            dist = proj.distance(point);
            index = i;
        }
    }
    if (index == -1){
        return Point(1000,10000,1000);
    }
    return maxisPoint[index];
}

float Landform::distanceToMedialAxis(Point p){
    float dist = 1000.0;
    
    vector<Point> maxisPoint = base.getSkeleton().getGeometricPoints();
    if (maxisPoint.size() == 0){
        base.calculateMedialAxis();
    }
    for (int j = 0; j < maxisPoint.size(); j++){
        dist = min(dist, p.distance(maxisPoint[j]));
    }
    if (dist > 100) printf("Erro no calculo da distancia ate o eixo medial!!!\n\n");
    return dist;
}


void Landform::calculateBox(){
    vector<Point> base_points = base.getPoints();
    vector<Point> silhuette_points = base.getPoints();
    vector<Point> maxis_points = base.getSkeleton().getGeometricPoints();
    
    for (int i = 0; i < silhuette_points.size(); i++){
        base_points.push_back(silhuette_points[i]);
    }
    for (int i = 0; i < maxis_points.size(); i++){
        base_points.push_back(maxis_points[i]);
    }
    
    box.resizeBox(base_points);
}


vector<Point> Landform::getMaxis(){
    return base.getSkeleton().getGeometricPoints();
}

vector< pair<Point, Point> > Landform::getNormals(){
    return normals;
}

void Landform::setupTerrainData(int r, int c, float w, float d){
    this->rows = r;
    this->cols = c;
    this->width = w;
    this->depth = d;
}


void Landform::setHeight(float _height){
    this->height = _height;
    colorize->setPeak(height);
}

float Landform::getHeight(){
    return this->height;
}
        
void Landform::addSilhouettePoint(Point p){
    silhouette.addPoint(p);
}

Trace Landform::getSilhouette(){
    return silhouette;
}

void Landform::addBasePoint(Point p){
    base.addPoint(p);
    copyBase.addPoint(p);
}

Base Landform::getBase(){
    return base;
}

void Landform::adjustSilhouette(){
    silhouette.adjust(true);
}

void Landform::adjustBase(){
    if (rows*cols*width*depth != 0 and base.size() > 100){
        projectBaseToMesh();
        meshBase = removeRedundancy(meshBase);
        completeBaseInMesh(meshBase);
    }
    printf("\nAdjust: base size  >>  %i  >>  %i  >> %i \n\n", 
                        base.size(), meshBase.size(), adjusted.size() );
}

float Landform::calculateBaseArea(){
    float area = 0.0;
    if (base.size() > 100){
        adjustBase();
        for (int i = 0; i < base.size()-1; i++){
            area += (1.0/4.0)*( (base[i+1].getZ() - base[i].getZ())
                                 *(base[i].getX() + base[i+1].getX())
                              - (base[i+1].getX() - base[i].getX())
                                 *(base[i].getZ() + base[i+1].getZ()) );
        }
        area += (1.0/4.0)*( (base[0].getZ() - base[-1].getZ())
                             *(base[-1].getX() + base[0].getX())
                          - (base[0].getX() - base[-1].getX())
                             *(base[-1].getZ() + base[0].getZ()) );
    }
    return area;
}

void Landform::reduceBase(){
    float dist;
    int index;
    reduced.clear();
    Point normal;
    int s = base.size();
    float cos;
    
    Point der;
    for (int i = 0; i < base.size(); i++){
        normal = base.calculateNormal(i);
    }
}

void Landform::draw(bool showBase, bool showSilhouette, bool showShadow){
    
    if (showBase and base.size() > 2){
        glLineWidth(100);
        glColor3f(0.0, 0.0, 1.0);
        glBegin(GL_LINE_STRIP);
        for (int i = 0; i < base.size(); i++){
            base[i].draw();
        }
        glEnd();
    }
    if (showSilhouette  and silhouette.size() > 2){
        glColor3f(1.0, 0.0, 0.0);
        glBegin(GL_LINE_STRIP);
        for (int i = 0; i < silhouette.size(); i++){
            silhouette[i].draw();
        }
        glEnd();
    }
    if (showShadow and silhouette.size() > 2){
        Point& s1 = silhouette[0];
        Point& s2 = silhouette.getPoints().back();
        glColor3f(1.0, 0.0, 0.0);
        glBegin(GL_LINE_STRIP);
            glVertex3f(s1.getX(), s1.getY()+0.01, s1.getZ());
            glVertex3f(s2.getX(), s2.getY()+0.01, s2.getZ());
        glEnd();
    }
}

void Landform::draw(int mode, int vision){
    if ((mode == CONTOUR or mode == BOX) and vision == UP and base.size() > 2){
//        glLineWidth(100);
        glPointSize(3);
        glColor3f(0.0, 0.0, 1.0);
        glBegin(GL_POINTS);
        for (int i = 0; i < base.size(); i++){
            base[i].draw();
        }
        glEnd();
    }
    if (mode == CONTOUR and vision == FRONT and silhouette.size() > 2){
        glColor3f(1.0, 0.0, 0.0);
        glBegin(GL_LINE_STRIP);
        for (int i = 0; i < silhouette.size(); i++){
            silhouette[i].draw();
        }
        glEnd();
    }
    if (mode == CONTOUR and vision == UP and silhouette.size() > 2){
        Point& s1 = silhouette[0];
        Point& s2 = silhouette.getPoints().back();
        glColor3f(1.0, 0.0, 0.0);
        glBegin(GL_LINE_STRIP);
            glVertex3f(s1.getX(), s1.getY()+0.01, s1.getZ());
            glVertex3f(s2.getX(), s2.getY()+0.01, s2.getZ());
        glEnd();
    }
    if (mode == BOX or mode == SKELETON){
        box.draw(1.0, 0.0, 0.0);
        float w_inc = box.width / float(dem.size() - 1);
        float d_inc = box.depth / float(dem.size() - 1);
        glPointSize(2);
        glColor3f(0.0, 0.0, 1.0);
        glBegin(GL_POINTS);
        for (int  i = 0; i < dem.size(); i++){
            for (int  j = 0; j < dem.size(); j++){
                (box.origin+Point(j*w_inc,dem[i][j],i*d_inc)).draw();
            }
        }
        glEnd();
    }
    if (mode == SKELETON){
        vector< Point > skelPoints = base.getSkeleton().getGeometricPoints();
        glPointSize(2);
        glColor3f(0.0, 0.0, 1.0);
        glBegin(GL_POINTS);
        for (int i = 0; i < skelPoints.size(); i++){
            skelPoints[i].draw();
        }
        glEnd();

        skelPoints = base.getSkeleton().getTopologicalPoints();
        for (int i = 0; i < skelPoints.size(); i++){
            glColor3f(0.0, 0.0, 1.0);
            Sphere s(skelPoints[i], 0.1100, 20+i);
            glLoadName(20+i);
            s.draw();
        }
        Sphere s(Point(0.0,0.0,0.0), 0.00100, 10000);
        glLoadName(1000);
        s.draw();
        
//        glColor3f(0.0, 1.0, 1.0);
//        for (int i = 0; i < maxisProj.size(); i++){
//            glBegin(GL_LINES);
//            glVertex3f(maxisProj[i].first.getX(), 0.0, maxisProj[i].first.getZ());
//            glVertex3f(maxisProj[i].second.getX(), 0.0, maxisProj[i].second.getZ());
//            glEnd();    
//        }
        
    }
}

//////////////////
//  HaroldHill  //
//////////////////

HaroldHill::HaroldHill(){
    colorize = new HillColorize();
}

HaroldHill::~HaroldHill(){

}

float HaroldHill::distance(Point point){
    Point proj = projection(point);
    proj.setY(point.getY());
    return (point - proj).lenght();
}


Point HaroldHill::projection(Point point){
    Point& first = silhouette.getPoints().front();
    Point& last = silhouette.getPoints().back();

    Point diff = last - first;

    Point p = point - first;
    Point proj = (diff * diff.dotProduct(p) ) / ( diff.dotProduct(diff) );
    proj += first;

    Point shad;
    float dist = 1000.0;
    float height = 0.0;
    
    
    for (int i = 0; i < silhouette.size(); i++){
        shad = Point(silhouette[i].getX(), 0.0, silhouette[i].getZ());
        if ( ( shad - proj).lenght() < dist ){
            dist = (shad - proj).lenght();
            height = silhouette[i].getY();
        }
    }
    
    proj.setY(height);   
    return proj;
}

float HaroldHill::kernelDist(float d){
    float kd = 1 - pow(d/3.0, 2);
    if (kd < 0){
        return 0.0;
    }
    else{
        return kd;
    }
}

int HaroldHill::getType(){
    return HAROLD_HILL;
}

float HaroldHill::getPointHeight(Point p){
    float weight = kernelDist( distance( p ) );
//        return ( (1 - weight) * p.getY() ) + ( weight * projection(p).getY() );
    return weight * projection(p).getY();
}

////////////////////
//  DistanceHill  //
////////////////////

DistanceHill::DistanceHill(){
    colorize = new HillColorize();
}

DistanceHill::~DistanceHill(){

}

int DistanceHill::getType(){
    return DISTANCE_HILL;
}

void DistanceHill::setSlopeType(int type){
    slope = type;
}

int DistanceHill::getSlopeType(){
    return slope;
}


void DistanceHill::finishBase(){
    base.finishBase();
}

float DistanceHill::getPointHeight(Point p){
    float h = 0.0;
    float d = 1000.0;           // distancia do ponto ate a base
    int index = -1;             // index do ponto na base

    p.setY(0.0);
    if (base.isInternalPoint(p)){
        d = 1000.0;
        for (int i = 0; i < base.size(); i++){
            if( (base[i] - p).lenght() < d ){
                d = (base[i] - p).lenght();
                index = i;
            }
        }
        h = height * d;
    }
    return h;
}

/////////////////////////
//  MaxisDistanceHill  //
/////////////////////////

MaxisDistanceHill::MaxisDistanceHill(){
    colorize = new HillColorize();
    res = 200;
}

MaxisDistanceHill::~MaxisDistanceHill(){

}

int MaxisDistanceHill::getType(){
    return MAXISDISTANCE_HILL;
}

void MaxisDistanceHill::setSlopeType(int type){
    slope = type;
}

int MaxisDistanceHill::getSlopeType(){
    return slope;
}


void MaxisDistanceHill::finishBase(){
    base.finishBase();
    createDem();
}

void MaxisDistanceHill::createDem(){
    calculateBox();
    float w_inc = box.width / float(res - 1);
    float d_inc = box.depth / float(res - 1);

    dem.clear();
    dem.resize(res);

    basePointer.clear();
    basePointer.resize(res);

    for (int i = 0; i < res; i++){
        dem[i].resize(res);
        basePointer[i].resize(res);
        for (int j = 0; j < res; j++){
            basePointer[i][j] = -1;
            dem[i][j] = getPointHeightDem(box.origin+Point(j*w_inc, 0.0, i*d_inc));
        }
    }
    for (int i = 0; i < res; i++){
        for (int j = 0; j < res; j++){
            dem[i][j] = base.getBaseMaxisDistance(basePointer[i][j]).first;
        }
    }
}


float MaxisDistanceHill::getPointHeightDem(Point p){
    float h = 0.0;
    float d;                // distancia do ponto ate a base
    int index;              // index do ponto na base

    p.setY(0.0);
    if (base.isInternalPoint(p)){
        index = base.projectPointBase(p);
        d = base[index].distance(p);
        base.updateBaseMaxisDistance(index, d);

        if (fabs(box.width*box.depth) < 0.01) calculateBox();
        if (p.getX() > box.origin.getX() and p.getX() < box.origin.getX() + box.width
        and p.getZ() > box.origin.getZ() and p.getZ() < box.origin.getZ() + box.depth){
            Point diff = p - box.origin;
            int col = int( diff.getX() * float(res) / box.width );
            int row = int( diff.getZ() * float(res) / box.depth );
            basePointer[row][col] = index;
        }
        h = d;
    }
    return h;
}


float MaxisDistanceHill::getPointHeight(Point p){
    float h = 0.0;

    if (base.isInternalPoint(p)){
        if (fabs(box.width*box.depth) < 0.01) calculateBox();
        if (p.getX() > box.origin.getX() and p.getX() < box.origin.getX() + box.width
        and p.getZ() > box.origin.getZ() and p.getZ() < box.origin.getZ() + box.depth){
            Point diff = p - box.origin;
            int col = int( diff.getX() * float(res - 1) / box.width );
            int row = int( diff.getZ() * float(res - 1) / box.depth );
            h = height * dem[row][col];
        }
    }
    return h;
}

//////////////////////
//  NormalizedHill  //
//////////////////////

NormalizedHill::NormalizedHill(){
    colorize = new HillColorize();
    res = 200;
}

NormalizedHill::~NormalizedHill(){

}

int NormalizedHill::getType(){
    return NORMALIZED_HILL;
}

void NormalizedHill::setSlopeType(int type){
    slope = type;
}

int NormalizedHill::getSlopeType(){
    return slope;
}


void NormalizedHill::finishBase(){
    base.finishBase();
    createDem();
}

void NormalizedHill::createDem(){
    calculateBox();
    float w_inc = box.width / float(res - 1);
    float d_inc = box.depth / float(res - 1);

    dem.clear();
    dem.resize(res);

    basePointer.clear();
    basePointer.resize(res);

    for (int i = 0; i < res; i++){
        dem[i].resize(res);
        basePointer[i].resize(res);
        for (int j = 0; j < res; j++){
            basePointer[i][j] = -1;
            dem[i][j] = getPointHeightDem(box.origin+Point(j*w_inc, 0.0, i*d_inc));
        }
    }

    for (int i = 0; i < res; i++){
        for (int j = 0; j < res; j++){
            dem[i][j] = 5.0 * dem[i][j] / base.getBaseMaxisDistance(basePointer[i][j]).first;
        }
    }
}


float NormalizedHill::getPointHeightDem(Point p){
    float h = 0.0;
    float d;                // distancia do ponto ate a base
    int index;              // index do ponto na base

    p.setY(0.0);
    if (base.isInternalPoint(p)){
        index = base.projectPointBase(p);
        d = base[index].distance(p);
        base.updateBaseMaxisDistance(index, d);

        if (fabs(box.width*box.depth) < 0.01) calculateBox();
        if (p.getX() > box.origin.getX() and p.getX() < box.origin.getX() + box.width
        and p.getZ() > box.origin.getZ() and p.getZ() < box.origin.getZ() + box.depth){
            Point diff = p - box.origin;
            int col = int( diff.getX() * float(res) / box.width );
            int row = int( diff.getZ() * float(res) / box.depth );
            basePointer[row][col] = index;
        }
        h = d;
    }
    return h;
}


float NormalizedHill::getPointHeight(Point p){
    float h = 0.0;

    if (base.isInternalPoint(p)){
        if (fabs(box.width*box.depth) < 0.01) calculateBox();
        if (p.getX() > box.origin.getX() and p.getX() < box.origin.getX() + box.width
        and p.getZ() > box.origin.getZ() and p.getZ() < box.origin.getZ() + box.depth){
            Point diff = p - box.origin;
            int col = int( diff.getX() * float(res - 1) / box.width );
            int row = int( diff.getZ() * float(res - 1) / box.depth );
            h = height * dem[row][col];
        }
    }
    return h;
}


//////////////////
//  StepPlateu  //
//////////////////

StepPlateu::StepPlateu(){
    colorize = new PlateuColorize();
}

StepPlateu::~StepPlateu(){

}

int StepPlateu::getType(){
    return STEP_PLATEU;
}

float StepPlateu::getPointHeight(Point p){
    if (base.isInternalPoint(p)){
        return height;
    }
    return 0.0;
}

//////////////////
//  RampPlateu  //
//////////////////

RampPlateu::RampPlateu(){
    colorize = new PlateuColorize();
}

RampPlateu::~RampPlateu(){

}

int RampPlateu::getType(){
    return RAMP_PLATEU;
}

void RampPlateu::finishBase(){
    base.finishBase();
}

float RampPlateu::getPointHeight(Point p){
    float h = 0.0;
    float d = 1000.0;           // distancia do ponto ate a base
    int index = -1;             // index do ponto na base

    
    p.setY(0.0);
    if (base.isInternalPoint(p)){
        d = 1000.0;
        for (int i = 0; i < base.size(); i++){
            if( (base[i] - p).lenght() < d ){
                d = (base[i] - p).lenght();
                index = i;
            }
        }
        h = d < height ? d : height;
    }
    return height;
}


////////////////////
//  SmoothPlateu  //
////////////////////

SmoothPlateu::SmoothPlateu(){
    colorize = new PlateuColorize();
}

SmoothPlateu::~SmoothPlateu(){

}

int SmoothPlateu::getType(){
    return SMOOTH_PLATEU;
}

float SmoothPlateu::getPointHeight(Point p){
    float h = 0.0;
    float d = 1000.0;           // distancia do ponto ate a base
    int index = -1;             // index do ponto na base

    p.setY(0.0);
    if (base.isInternalPoint(p)){
        d = 1000.0;
        for (int i = 0; i < base.size(); i++){
            if( (base[i] - p).lenght() < d ){
                d = (base[i] - p).lenght();
                index = i;
            }
        }
//        h = d < 2.0 ? 3.0 * smooth( 0.0, 2.0, d) : 3.0;
        h = d < slopeInc ? height * smooth( 0.0, slopeInc, d) : height;
    }
    return h;
}



///////////////////
//  MaxisPlateu  //
///////////////////

MaxisPlateu::MaxisPlateu(){
    res = 200;
    colorize = new PlateuColorize();
}

MaxisPlateu::~MaxisPlateu(){

}

int MaxisPlateu::getType(){
    return MAXIS_PLATEU;
}

void MaxisPlateu::finishBase(){
    base.finishBase();
    createDem();
}

void MaxisPlateu::createDem(){
    calculateBox();
    float w_inc = box.width / float(res - 1);
    float d_inc = box.depth / float(res - 1);

    dem.clear();
    dem.resize(res);

    basePointer.clear();
    basePointer.resize(res);

    for (int i = 0; i < res; i++){
        dem[i].resize(res);
        basePointer[i].resize(res);
        for (int j = 0; j < res; j++){
            basePointer[i][j] = -1;
            dem[i][j] = getPointHeightDem(box.origin+Point(j*w_inc, 0.0, i*d_inc));
        }
    }
    for (int i = 0; i < res; i++){
        for (int j = 0; j < res; j++){
            float D = base.getBaseMaxisDistance(basePointer[i][j]).first;
            float d = dem[i][j];
//            dem[i][j] = (d < max(D-1.0, 0.0) ? (d*3.0)/(D-1) : 3.0);
            if (D > slopeInc)
                dem[i][j] = (d/D < 0.6 ? (height/0.6)*(d/D) : height);
            else
                dem[i][j] = min(d * height / (slopeInc * 0.6f), height);
        }
    }
}


float MaxisPlateu::getPointHeightDem(Point p){
    float h = 0.0;
    float d;                // distancia do ponto ate a base
    int index;              // index do ponto na base

    p.setY(0.0);
    if (base.isInternalPoint(p)){
        index = base.projectPointBase(p);
        d = base[index].distance(p);
        base.updateBaseMaxisDistance(index, d);

        if (fabs(box.width*box.depth) < 0.01) calculateBox();
        if (p.getX() > box.origin.getX() and p.getX() < box.origin.getX() + box.width
        and p.getZ() > box.origin.getZ() and p.getZ() < box.origin.getZ() + box.depth){
            Point diff = p - box.origin;
            int col = int( diff.getX() * float(res) / box.width );
            int row = int( diff.getZ() * float(res) / box.depth );
            basePointer[row][col] = index;
        }
        h = d;
    }
    return h;
}

float MaxisPlateu::getPointHeight(Point p){
    float height = 0.0;

    if (base.isInternalPoint(p)){
        if (fabs(box.width*box.depth) < 0.01) calculateBox();
        if (p.getX() > box.origin.getX() and p.getX() < box.origin.getX() + box.width
        and p.getZ() > box.origin.getZ() and p.getZ() < box.origin.getZ() + box.depth){
            Point diff = p - box.origin;
            int col = int( diff.getX() * float(res - 1) / box.width );
            int row = int( diff.getZ() * float(res - 1) / box.depth );
            height = dem[row][col];
        }
    }
    return height;
}


///////////////
//  Volcano  //
///////////////

Volcano::Volcano(){
    colorize = new VolcanoColorize();
    res = 200;
}

Volcano::~Volcano(){

}

int Volcano::getType(){
    return VOLCANO;
}


void Volcano::finishBase(){
    base.finishBase();
    createDem();
}

void Volcano::createDem(){
    float maxim = 0.0;

    calculateBox();
    w_inc = box.width / float(res - 1);
    d_inc = box.depth / float(res - 1);

    dem.clear();
    dem.resize(res);
    for (int i = 0; i < res; i++){
        dem[i].resize(res);
        for (int j = 0; j < res; j++){
            dem[i][j] = height * getPointHeightDem(box.origin+Point(j*w_inc, 0.0, i*d_inc));
            if (dem[i][j] > maxim) maxim = dem[i][j];
        }
    }

    float cut = maxim * (slopeInc + 1.0) /6.0f;
    for (int i = 0; i < res; i++){
        for (int j = 0; j < res; j++){
            if (dem[i][j] >= cut)
                dem[i][j] = max(2 * cut - dem[i][j], 0.0f);
        }
    }

    colorize->setPeak(0.9*cut);
}


float Volcano::getPointHeightDem(Point p){
    float h = 0.0;
    p.setY(0.0);
    if (base.isInternalPoint(p)){
        h = base[base.projectPointBase(p)].distance(p);
    }
    return h;
}


float Volcano::getPointHeight(Point p){
    float height = 0.0;
    if (base.isInternalPoint(p)){
        if (d_inc*w_inc == 0){
            //calculateBox();
            w_inc = box.width / float(res - 1);
            d_inc = box.depth / float(res - 1);
        }
        Point diff = p - box.origin;
        height = dem[int( diff.getZ() / d_inc)][int( diff.getX() / w_inc)];
    }
    return height;
}


//////////////
//  Island  //
//////////////

Island::Island(){
    colorize = new IslandColorize();
}

Island::~Island(){

}

int Island::getType(){
    return ISLAND;
}

void Island::finishBase(){
    base.finishBase();
}

float Island::getPointHeight(Point p){
    float h = 0.0;
    float d = 1000.0;           // distancia do ponto ate a base
    int index = -1;             // index do ponto na base

    p.setY(0.0);
    if (base.isInternalPoint(p)){
        d = 1000.0;
        for (int i = 0; i < base.size(); i++){
            if( (base[i] - p).lenght() < d ){
                d = (base[i] - p).lenght();
                index = i;
            }
        }

        if (d < slopeInc + sin(p.getX()) * cos(p.getZ() + p.getX()))
            h = d / slopeInc;
        else
            h = max(min(slopeInc, 1.0f), fabs( fabs(height-slopeInc)*d*sin(p.getX())/2.0f + cos(p.getZ())/2.0f));
    }
    return h;
}

////////////
//  Flat  //
////////////

Flat::Flat(){

}

Flat::~Flat(){

}

int Flat::getType(){
    return FLAT;
}

float Flat::getPointHeight(Point p){
    if (base.isInternalPoint(p)){
        return this->height;
    }
    return 0.0;
}


//////////////////
//  Depression  //
//////////////////

Depression::Depression(){

}

Depression::~Depression(){

}

int Depression::getType(){
    return DEPRESSION;
}

float Depression::getPointHeight(Point p){
    if (base.isInternalPoint(p)){
        return this->height;
    }
    return 0.0;
}
