#ifndef LANDFORM_H
#define LANDFORM_H

#include <vector>

#include "trace.h"
#include "terraincontours.h"
#include "maxis.h"
#include "functions.h"
#include "texture.h"

#define BASE        1
#define SILHOUETTE  2

#define HAROLD_HILL             11
#define DISTANCE_HILL           12
#define NORMALIZED_HILL         13
#define MAXISDISTANCE_HILL      14
#define STEP_PLATEU             15
#define RAMP_PLATEU             16
#define SMOOTH_PLATEU           17
#define MAXIS_PLATEU            18
#define VOLCANO                 19
#define ISLAND                  20
#define FLAT                    21
#define DEPRESSION              22

class Landform{
    protected:
        int rows; 
        int cols; 
        float width; 
        float depth;
        float height;
        float slopeInc;

        int slope;
        
        Trace silhouette;
        Base base;
        Base copyBase;
        
        Trace meshBase;
        Trace adjusted;
        Trace reduced;
        Trace maxis;
        Trace meshMaxis;

        
        vector< vector<float> > dem;
        vector< vector<int> > basePointer;
        
        vector< pair<Point, Point> > normals;
        vector< pair<Point, Point> > maxisProj;
        
        Box box;
        
        Point getMeshDomainPoint(int i, int j);
        void projectBaseToMesh();
        void projectMaxisToMesh();
        Trace removeRedundancy(Trace trace);
        void completeBaseInMesh(Trace trace);
        
        Point baseProjection(Point point, Point normal);
        float distanceToMedialAxis(Point p);
        
        void calculateBox();


        
    public:
        LandColorize* colorize;

        Landform();
        ~Landform();
        
        vector<Point> getMaxis();
        vector< pair<Point, Point> > getNormals();
        
        void setupTerrainData(int r, int c, float w, float d);
        
        virtual int getType(){}
        
        void setHeight(float _height);
        float getHeight();

        void setSlopeInc(float s){slopeInc = s;}
        float getSlopeInc(){return slopeInc;}
        
        void addSilhouettePoint(Point p);
        Trace getSilhouette();

        void addBasePoint(Point p);
        Base getBase();
        
        virtual void finishBase(){}

        void adjustSilhouette();
        void adjustBase();
        float calculateBaseArea();
        void reduceBase();

        virtual float getPointHeight(Point p){}
        
        void draw(bool showBase, bool showSilhouette, bool showShadow);
        void draw(int mode, int vision);
};

class HaroldHill : public Landform{
    private:
        float distance(Point point);
        Point projection(Point point);
        float kernelDist(float d);
        
    public:
        HaroldHill();
        ~HaroldHill();
        
        int getType();

        float getPointHeight(Point p);
};

class DistanceHill : public Landform{
    public:
        DistanceHill();
        ~DistanceHill();

        int getType();

        void setSlopeType(int type);
        int getSlopeType();

        void finishBase();
        float getPointHeight(Point p);
};

class MaxisDistanceHill : public Landform{
    private:
        int res;

    public:
        MaxisDistanceHill();
        ~MaxisDistanceHill();

        int getType();

        void setSlopeType(int type);
        int getSlopeType();

        float getPointHeightDem(Point p);
        void createDem();

        void finishBase();
        float getPointHeight(Point p);
};

class NormalizedHill : public Landform{
    private:
        int res;

    public:
        NormalizedHill();
        ~NormalizedHill();

        int getType();

        void setSlopeType(int type);
        int getSlopeType();

        float getPointHeightDem(Point p);
        void createDem();

        void finishBase();
        float getPointHeight(Point p);
};


class StepPlateu : public Landform{
    public:
        StepPlateu();
        ~StepPlateu();
        
        int getType();

        float getPointHeight(Point p);
};

class RampPlateu : public Landform{
    public:
        RampPlateu();
        ~RampPlateu();
        
        int getType();
        
        void finishBase();
        float getPointHeight(Point p);
};

class SmoothPlateu : public Landform{
    public:
        SmoothPlateu();
        ~SmoothPlateu();
        
        int getType();

        float getPointHeight(Point p);
};

class MaxisPlateu : public Landform{
private:
    int res;

    public:
        MaxisPlateu();
        ~MaxisPlateu();

        int getType();

        float getPointHeightDem(Point p);
        void createDem();

        void finishBase();
        float getPointHeight(Point p);
};

class Volcano : public Landform{
    private:
        Point center;
        int res;
        float w_inc, d_inc;

    public:
        Volcano();
        ~Volcano();

        int getType();

        float getPointHeightDem(Point p);
        void createDem();

        void finishBase();
        float getPointHeight(Point p);
};

class Island : public Landform{
    private:
        Point center;
        int res;

    public:
        Island();
        ~Island();

        int getType();

        void finishBase();
        float getPointHeight(Point p);
};

class Flat : public Landform{
    public:
        Flat();
        ~Flat();
        
        int getType();

        float getPointHeight(Point p);
};

class Depression : public Landform{
    public:
        Depression();
        ~Depression();
        
        int getType();

        float getPointHeight(Point p);
};

#endif // LANDFORM_H
