#ifndef SKELETON_H
#define SKELETON_H

#include <cstdio>
#include <vector>
#include <map>

#include "primitive.h"

using namespace std;

class TopologicalVertex;

class GeometricVertex{
    private:
        Point position;
        TopologicalVertex* vertex1;
        TopologicalVertex* vertex2;
        
        float height;
        
    public:
        GeometricVertex();
        GeometricVertex(Point pos);
        ~GeometricVertex();
        
        void setTopologicalVertices(TopologicalVertex* top1, TopologicalVertex* top2);
        pair<TopologicalVertex*,TopologicalVertex*> getTopologicalVertex();
        
        void setPosition(Point pos);
        Point getPosition();

        void setHeight(float h);
        float getHeight();
        
        void show();
        
        bool operator==(Point vertex);
        bool operator==(GeometricVertex* vertex);
};


class TopologicalVertex{
    private:
        GeometricVertex* vertex;
        vector< vector<GeometricVertex*> > internalVertices;
        float height;
        int geoIndex;
    
    public:
        TopologicalVertex();
        TopologicalVertex(GeometricVertex* v);
        ~TopologicalVertex();
        
        int size();
        
        void setHeight(float h);
        float getHeight();
        
        void setPosition(Point pos);
        Point getPosition();
        
        void setGeometricIndex(int index){geoIndex = index;}
        int getGeometricIndex(){return geoIndex;}
        
        GeometricVertex* getVertex();
        vector< vector<GeometricVertex*> > getInternalVertices();
        vector<Point> getAdjacentVertices();
        
        void removeBranch(int i);        
        void addBranch(vector<GeometricVertex*> branch);
        vector<GeometricVertex*>& findBranch(GeometricVertex* vertex);
        int getIndexBranch(GeometricVertex* vertex);
        
        void show();
        
        bool operator==(GeometricVertex* v);
        bool operator==(TopologicalVertex* v);
};

class GeometricEdge{
    private:
        // mudar para ponteiros que referenciam GeometricVertex
        pair<int, int> vertices;
        
    public:
        GeometricEdge();
        GeometricEdge(int v1, int v2);
        ~GeometricEdge(){}
        
        void setVertex1Index(int index);
        int getVertex1Index();

        void setVertex2Index(int index);
        int getVertex2Index();
        
        pair<int, int> getVerticesIndex();
        
        bool operator==(GeometricEdge edge);
};

class TopologicalEdge{
    private:
        TopologicalVertex* vertex1;
        TopologicalVertex* vertex2;
        vector<GeometricVertex*> internalVertices;
        vector<float> distances;
        
    public:
        TopologicalEdge();
        TopologicalEdge(TopologicalVertex* vertex1, TopologicalVertex* vertex2);
        ~TopologicalEdge();
        
        void setTopologicalVertex1(TopologicalVertex* v1);
        TopologicalVertex* getTopologicalVertex1();

        void setTopologicalVertex2(TopologicalVertex* v2);
        TopologicalVertex* getTopologicalVertex2();
        
        pair<TopologicalVertex*, TopologicalVertex*> getVerticesIndex();
        
        void addInternalVertex(GeometricVertex* geo, bool back = true);
        vector<GeometricVertex*> getInternalVertices();

        void updateDistances();
        void updateGeometricVertexHeight();
        
        bool operator==(TopologicalEdge edge);
        void show();
};

class GeometricSkeleton{
    private:
        map<int, int> markup;
        map<int, GeometricVertex*> vertices;
        vector<Point> points;
        vector< GeometricEdge > edges;
        
    public:
        GeometricSkeleton();
        ~GeometricSkeleton();
        
        void mark();
        
        GeometricVertex* getVertex(int i);
        map<int, GeometricVertex*> getVertices();
        int verticesSize();
        void deleteVertex(int index);
        void removeUnusedVertices();
        void removeNullVertices();
        map<int, int> markVertices();
        
        int edgesSize();
        GeometricEdge getEdge(int index);
        void addEdge(Point v1, Point v2);
        void deleteEdge(int index);

        void removeSmallEdges();
        
        int findVertexIndex(GeometricVertex* v);
        vector<int> findEdge(int index);
        int findEdge(GeometricVertex* v1, GeometricVertex* v2, bool show = false);
        
        void updateGeometricPoints();
        Point getPoint(int i);
        vector<Point> getGeometricPoints();
        
        void showVertices();
        void showEdges();
};

class TopologicalSkeleton{
    private:
        map<int, int> markup;
        vector<TopologicalVertex*> vertices;
        vector<TopologicalEdge> edges;
        
        vector<Point>  verticesPoints;
        vector< pair<Point,Point> > edgesPoints;
        
        GeometricSkeleton& geometric;
        
    public:
        TopologicalSkeleton(GeometricSkeleton& geometric);
        ~TopologicalSkeleton();
        
        int verticesSize();
        TopologicalVertex* getVertex(int index);
        vector<TopologicalVertex*> getVertices();
        vector<Point> getVerticesPoints();
        void addVertex(TopologicalVertex* vertex);
        void deleteVertex(int topIndex);
        void createVertex(int topIndex);
        void selectVertex();
        int findVertex(TopologicalVertex* v);
        int findVertex(GeometricVertex* v);
        
        void verticesSimplification();
        
        int edgesSize();
        TopologicalEdge& getEdge(int index);
        vector<TopologicalEdge> getEdges();
        vector< pair<Point,Point> > getEdgesPoints();
        int findEdge(GeometricVertex* v1, GeometricVertex* v2);
        void updateEdges();

        void markByEdge();
        
        void showVertices();
        void showEdges();
};

class Skeleton{
    private:
        GeometricSkeleton geometric;
        TopologicalSkeleton topologic;
        
        void complete();
        
    public:
        Skeleton();
        ~Skeleton();
        
        GeometricSkeleton& getGeometricSkeleton();
        TopologicalSkeleton& getTopologicalSkeleton();

        int size();
        Point operator[](int i);
        vector<Point> getGeometricPoints();
        GeometricVertex* getGeometricVertex(int index);
        vector<Point> getTopologicalPoints();
        vector< pair<Point,Point> > getTopologicalEdgesPoints();
        
        void updateGeometricHeights();
        
        void addEdge(Point v1, Point v2);
        void finishSkeleton();
        void show();
        void edgeShow();
};

#endif // SKELETON_H
