#include "terrain.h"
#include "detail.h"

#include <cmath>

void TerrainIO::saveData(string filename, Terrain* t){

}

void TerrainIO::readData(string filename, Terrain* t){

}

void TerrainIO::show(Terrain* t){

}


Terrain::Terrain() : mesh(&grid){
    width = 1.0;
    depth = 1.0;

    init();
}

Terrain::Terrain(float w, float d) : 
    mesh(&grid), 
    grid(Point(-w/2.0, 0.0, -d/2.0), w, d)
{
    width = w;
    depth = d;

    init();
}

Terrain::~Terrain(){

}

void Terrain::init(){
    modeType = POLYGONS;
    height = 0.0;

    finish_base = false;
    finish_silhouette = false;

    level = 1;
    camera = Point(0.0, 0.0, 0.0);
    grid.setCamera(camera);

    grid = Dem( Point(0.0,0.0,0.0), 1.0, 1.0);
    updateGridLevel(1);
    updateGridLevel(-1);

    colorize = false;
    showMesh = false;
    noise = false;
    bumpmapping = false;

    landHeight = 0.5;
    landSlope = 0.5;
}

void Terrain::updateGridLevel(int l){
    grid.change_level(l);
    mesh.adapt_simplify(0);
    mesh.adapt_refine(0);
}

void Terrain::changeLevel(int l){
    if (level >= 0 and level <= 20){
        level += l;
        if (level <= 0)
            level = 1;
        else if (level > 20)
            level = 20;
        else{
            printf("\nLevel: %i (ter)\n",level);fflush(stdout);
            updateGridLevel(l);
        }
    }
}

void Terrain::changeCamera(Point t, Point r){
    camera.setX(0.0);
    camera.setY(0.0);
    camera.setZ(0.0);

    camera.translate(t);

    r *= -3.14159265358979323 / 180.0;
    camera.rotate(0, r.getX());
    camera.rotate(1, r.getY());
    camera.rotate(2, r.getZ());

    grid.setCamera(camera);
    updateGridLevel(0);
}

Point Terrain::getCamera(){
    return camera;
}


void Terrain::clear(){
    landforms.clear();
    generatePlane(rows, cols);
}


void Terrain::plan(){
    generatePlane(rows, cols);
}

void Terrain::setShowMesh(bool show){
    showMesh = show;
}

bool Terrain::getShowMesh(){
    return showMesh;
}

void Terrain::setColorize(bool color){
    colorize = color;
}

bool Terrain::getColorize(){
    return colorize;
}

void Terrain::setNoise(bool n){
    noise = n;
    updateSurface();
    drawMesh();
}

bool Terrain::getNoise(){
    return noise;
}

void Terrain::setBumpMapping(bool bump){
    bumpmapping = bump;
    grid.setBumpMapping(bump);
    updateSurface();
    drawMesh();
}

bool Terrain::getBumpMapping(){
    return bumpmapping;
}

vector< vector<float> >& Terrain::getDem(){
    return dem;
}

Landform* Terrain::createLandform(int type){
    Landform* land = NULL;
    switch (type){
    case HAROLD_HILL:
        land = new HaroldHill();
        break;
    case DISTANCE_HILL:
        land = new DistanceHill();
        break;
    case MAXISDISTANCE_HILL:
        land = new MaxisDistanceHill();
        break;
    case NORMALIZED_HILL:
        land = new NormalizedHill();
        break;
    case STEP_PLATEU:
        land = new StepPlateu();
        break;
    case RAMP_PLATEU:
        land = new RampPlateu();
        break;
    case SMOOTH_PLATEU:
        land = new SmoothPlateu();
        break;
    case MAXIS_PLATEU:
        land = new MaxisPlateu();
        break;
    case VOLCANO:
        land = new Volcano();
        break;
    case ISLAND:
        land = new Island();
        break;
    }
    if (land){
        this->landforms.push_back(land);
        this->land_mark.push_back(false);
    }
    land->setupTerrainData(rows, cols, width, depth);
    land->setHeight(landHeight);
    land->setSlopeInc(landSlope);
    return land;
}


void Terrain::startSilhouette(){
    finish_silhouette = false;
    silhouette.clear();
}
void Terrain::addSilhouettePoint(Point p){
    if (! finish_silhouette)
        silhouette.addPoint(p);
}

Trace Terrain::getSilhouette(){
    return silhouette;
}

void Terrain::finishSilhouette(){
    finish_silhouette = true;
}

void Terrain::adjustSilhouette(){
    silhouette.adjust(true);
}


void Terrain::startBase(){
    finish_base = false;
    base.clear();
}
void Terrain::addBasePoint(Point p){
    if (!finish_base){
        base.addPoint(p);
    }
    else{
        printf("\n\nWARNING!!!\n");p.show(); 
        printf("POINT NOT ADDED!!!\n\n");
    }
}

Base& Terrain::getBase(){
    return base;
}

void Terrain::finishBase(){
    finish_base = true;

    base.smooth(10);
    base.filter(0.1);
    base.upsample(10);
    base.smooth(5);
    base.filter(0.01);


//    base.smooth(10);
//    base.filter(0.1);
//    base.upsample(10);
//    base.smooth(300);
//    base.filter(0.1);
}

void Terrain::addLandform(Landform* landform){
    landform->setupTerrainData(rows, cols, width, depth);
    this->landforms.push_back(landform);
    this->land_mark.push_back(false);
}

Landform* Terrain::getLandform(int index){
    if (index >= 0 and index < landforms.size()){
        return landforms[index];
    }
    else if ( landforms.size() > 0){
        return landforms.back();
    }
    else NULL;
}

Landform* Terrain::getLandform(Point p){
    Base landBase;
    for (int i = 0; i < landforms.size(); i++){
        landBase = landforms[i]->getBase();
        if ( landBase.isInternalPoint(p) ){
            return landforms[i];
        }
    }
    return NULL;
}

void Terrain::removeLandform(int index){
    if (index >= 0 and index < landforms.size()){
        printf("last\n");
        landforms.erase( landforms.begin() +  index );
    }
    else if ( landforms.size() > 0){
        printf("index: %i\n", index);
        landforms.erase( landforms.begin() +  landforms.size()-1 );
    }
    printf("no index: %i %i\n", landforms.size(), index);
}


void Terrain::setDimensions(float _width, float _depth){
    this->width = _width;
    this->depth = _depth;
    grid = Dem(Point(-width/2.0, 0.0, -depth/2.0), this->width, this->depth);
}

float Terrain::getPeak(){
    float peak = 0;
    
    for (int i = 0; i < this->rows; i++){
        for (int j = 0; j < this->rows; j++){
            peak = max(peak, this->dem[i][j]);
        }
    }
    this->height = peak;
    return peak;
}

void Terrain::readData(string filename, bool show){
    FILE* arq = fopen(filename.data(), "r");
    size_t len= 100;
    char* linha = (char*) malloc(len);
    int lands;
    int b_pts;
    int s_pts;
    int type;
    int slope;
    float height;
    
    fscanf(arq,"%[^\n]%*c\n", linha);
    if (show) printf("%s\n", linha);
    fscanf(arq,"%[^\n]%*c\n", linha);
    if (show) printf("\n%s\n",linha);
    fscanf(arq,"%s %f\n", linha, &(this->width));
    if (show) printf("width: %f\n", this->width);
    fscanf(arq,"%s %f\n", linha, &(this->depth));
    if (show) printf("depth: %f\n", this->depth);    
    fscanf(arq,"%s %f\n", linha, &(this->height));
    if (show) printf("height: %f\n", this->height);
    
    fscanf(arq, "%s %i\n", linha, &b_pts);
    if (show) printf("Base: %i\n", b_pts);
    for (int j = 0; j < b_pts; j++){
        float x, y, z;
        fscanf(arq, "%f  %f  %f\n", &x, &y, &z);
        base.addPoint(Point(x, y, z));
        if (show) printf("%f  %f  %f\n",base[j].getX(),base[j].getY(),base[j].getZ());
    }
    
    fscanf(arq, "%[^\n]%*c\n", linha);
    if (show) printf("\n%s\n",linha);
    fscanf(arq, "%s %i\n", linha, &lands);
    if (show) printf("%s %i\n", linha, lands); 

    for (int i = 0; i < lands; i++){
        fscanf(arq,"%[^\n]%*c\n",linha);
        if (show) printf("\n%s\n",linha);
        
        fscanf(arq, "%s %i\n", linha, &type);
        if (show) printf("%s %i\n", linha, type);
        Landform* landform = createLandform(type);
        
        fscanf(arq, "%s %f\n", linha, &height);
        if (show) printf("%s %f\n", linha, height);
        landform->setHeight(height);
        
        fscanf(arq, "%s %i\n", linha, &b_pts);
        if (show) printf("%s %i\n", linha, b_pts); 
        for (int j = 0; j < b_pts; j++){
            float x, y, z;
            fscanf(arq, "%f  %f  %f\n", &x, &y, &z);
            if (show) printf("%f  %f  %f\n",x, y, z);
            landform->addBasePoint( Point(x, y, z) );
        }
        
        fscanf(arq, "%s %i\n", linha, &s_pts);
        if (show) printf("%s %i\n", linha, s_pts); 
        for (int j = 0; j < s_pts; j++){
            float x, y, z;
            fscanf(arq, "%f  %f  %f\n", &x, &y, &z);
            if (show) printf("%f  %f  %f\n",x, y, z);
            landform->addSilhouettePoint( Point(x, y, z) );
        }
        addLandform(landform);
    }
    
    fscanf(arq,"%[^\n]%*c\n",linha);
    if (show) printf("\n%s\n",linha);
    fscanf(arq,"%s %i\n", linha, &(this->rows));
    if (show) printf("rows: %i\n", this->rows);
    fscanf(arq,"%s %i\n", linha, &(this->cols));
    if (show) printf("cols: %i\n", this->cols);    
    
    for (int i = 0; i < this->rows; i++){
        for (int j = 0; j < this->cols; j++){
            fscanf(arq, "%f  ",&dem[i][j]);
        }
    }
    
    fclose(arq);
    updateTerrain();
}

void Terrain::writeData(string filename){
    FILE* arq = fopen(filename.data(), "w+");
    size_t len= 100;
    char* linha = (char*) malloc(len);

    fprintf(arq,"# Sketch-based Terrain Modeling\n");
    
    fprintf(arq,"\n# dimensions\n");
    fprintf(arq,"width %f\n",this->width);
    fprintf(arq,"depth %f\n",this->depth);
    fprintf(arq,"height %f\n",this->getPeak());
    
    fprintf(arq,"\nBase %i\n", base.size());
    for (int j = 0; j < base.size(); j++){
        fprintf(arq,"%f  %f  %f\n",base[j].getX(),base[j].getY(),base[j].getZ());
    }
    
    
    fprintf(arq,"\n# Landforms\n");
    fprintf(arq,"quantity %i\n",landforms.size());
    for (int i = 0; i < landforms.size(); i++){
        fprintf(arq,"\n%ith Landform\n",i);
        fprintf(arq,"type: %i\n",landforms[i]->getType());
        fprintf(arq,"height: %f\n",landforms[i]->getHeight());
                
        Trace b = landforms[i]->getBase();
        fprintf(arq,"Base %i\n", b.size());
        for (int j = 0; j < b.size(); j++){
            fprintf(arq,"%f  %f  %f\n",b[j].getX(),b[j].getY(),b[j].getZ());
        }

        Trace s = landforms[i]->getSilhouette();
        fprintf(arq,"Silhouette %i\n", s.size());
        for (int j = 0; j < s.size(); j++){
            fprintf(arq,"%f  %f  %f\n",s[j].getX(),s[j].getY(),s[j].getZ());
        }
    }

    fprintf(arq, "\n# Digital Elevation Models\n");
    fprintf(arq, "rows %i\n", this->rows);
    fprintf(arq, "columns %i\n", this->cols);
    
    for (int i = 0; i < this->rows; i++){
        for (int j = 0; j < this->cols; j++){
            fprintf(arq, "%f  ",dem[i][j]);
        }
    }
    
    fclose(arq);
}


void Terrain::showData(){
    printf("Object  SHOW\n");
    for (int i = 0; i < dem.size(); i++){
        printf("\n");
        for (int j = 0; j < dem.size(); j++){
            printf("%.2f  ",dem[i][j]);
        }
    }
    printf("\n");
}

void Terrain::generatePlane(int r, int c){
    this->rows = r;
    this->cols = c;
    
    dem.clear();
    colors.clear();
    dem.resize(this->rows);
    colors.resize(this->rows);
    for (int i = 0; i < this->rows; i++){
        dem[i].resize(this->cols);
        colors[i].resize(this->cols);
        for (int j = 0; j < this->cols; j++){
            dem[i][j] = 0.0;
//            colors[i][j] = Point(0, 51, 102);
            colors[i][j] = Point(51, 51, 0);
        }
    }
}


void Terrain::updateTerrain(){
    float widthInc = this->width/this->rows;
    float depthInc = this->depth/this->cols;

    generatePlane(rows,cols);

    for (int k = 0; k < this->landforms.size() and landforms.size() > 0; k++){
        if (!land_mark[k]){
            landforms[k]->finishBase();
            land_mark[k] = true;
            printf("\n\nK: %i - %i\n",k,landforms.size());
            Base b = landforms[k]->getBase();
            if (b.size() > 5){
                Trace t;
                t.setPoints(b.getPoints());
                contour.addTrace(t);
                printf("Adding base: %i\n",b.size());
            }
            Skeleton skel = landforms[k]->getBase().getSkeleton();
            if (skel.size() > 5){
                Trace t;
                t.setPoints(skel.getGeometricPoints());
                contour.addTrace(t);
                printf("Adding skel: %i\n",skel.size());
            }
            else{
                printf("Skel not added: %i\n",skel.size());
            }
        }

        for (int i = 0; i < this->rows; i++){
            for (int j = 0; j < this->cols; j++){
                Point p( (-this->width/2) + (i*widthInc), this->dem[i][j], 
                         (-this->depth/2) + (j*depthInc));
                float h = landforms[k]->getPointHeight(p);

                if (dem[i][j] < h){
                    this->dem[i][j] = h;
                    if (landforms[k]->getBase().isInternalPoint(p)){
                        p.setY(this->dem[i][j]);
                        this->colors[i][j] = landforms[k]->colorize->getColor(p);
                    }
                }
            }
        }
    }

    updateSurface();
}

vector< vector<float> > Terrain::addNoise(){
    Detail detail;
    vector< vector<float> > noise_dem1 = detail.noise(rows, cols, -0.1, 0.1);
    vector< vector<float> > noise_dem2 = detail.noise(rows, cols, -0.1, 1.0);
    vector< vector<float> > noise_res;
    noise_res.resize(rows);

    for (int i = 0; i < this->rows; i++){
        for (int j = 0; j < this->rows; j++){
            if(dem[i][j] > 0.1)
                noise_res[i].push_back(dem[i][j] + noise_dem1[i][j]);
            else
                noise_res[i].push_back(dem[i][j] + noise_dem2[i][j]);
        }
    }
    return noise_res;
}

void Terrain::updateSurface(){
    grid.setBasic(true);
    mesh.adapt_simplify(0);
    grid.setBasic(false);

    if (noise)
        grid.setDem(addNoise());
    else
        grid.setDem(dem);

    grid.setColors(colors);
    grid.setContour(contour);
    grid.setDimensions(width, depth);
}

void Terrain::enablePlane(Point center, Point normal){
    showPlane = true;
    plane.setCenter(center); 
    plane.setNormal(normal);
}

void disablePlane(){

}


void Terrain::setMode(int m){
    this->modeType = m;
}

int Terrain::getMode(){
    return this->modeType;
}

void Terrain::switchMode(){
    switch(this->modeType){
        case POLYGONS:
            glBegin(GL_POLYGON);
            break;
        case POINTS:
            glBegin(GL_POINTS);
            break;
        case LINES:
            glBegin(GL_LINE_LOOP);
            break;
        case CONTOUR:
            glBegin(GL_LINE_STRIP);
            break;
        case BOX:
            glBegin(GL_LINE_STRIP);
            break;
        case SKELETON:
            glBegin(GL_LINE_STRIP);
            break;
        case CONE:
//            glBegin(GL_POINTS);
            break;
//        case SHADING:
////            glBegin(GL_LINE_STRIP);
//            break;
    }
}

void Terrain::draw(){
    draw(true, true, true, true);
}

void Terrain::draw(bool showBase, bool showSilhouette, bool showShadow, bool showBox){
    float widthInc = this->width/rows;
    float depthInc = this->depth/cols;
    
    glColor3f(0.0, 1.0, 0.0);
    glLineWidth(1);
    glPointSize(2);
    glLoadName(17);
    for (int i = 0; i < rows; i++){
        for (int j = 0; j < cols; j++){
            switchMode();
                glVertex3f( (-this->width/2) + (i*widthInc), 
                            dem[i][j], 
                            (-this->depth/2) + (j*depthInc) );
                            
                glVertex3f( (-this->width/2) + ((i+1)*widthInc), 
                            dem[i == rows-1? i : i+1][j],
                            (-this->depth/2) + (j*depthInc) );
                            
                glVertex3f( (-this->width/2) + ((i+1)*widthInc), 
                            dem[i == rows-1? i : i+1][j == cols-1? j : j+1],
                            (-this->depth/2) + ((j+1)*depthInc) );
            glEnd();
            switchMode();
                glVertex3f( (-this->width/2) + (i*widthInc), 
                            dem[i][j], 
                            (-this->depth/2) + (j*depthInc) );
                            
                glVertex3f( (-this->width/2) + (i*widthInc), 
                            dem[i][j == cols-1? j : j+1],
                            (-this->depth/2) + ((j+1)*depthInc) );
                            
                glVertex3f( (-this->width/2) + ((i+1)*widthInc), 
                            dem[i == rows-1? i : i+1][j == cols-1? j : j+1],
                            (-this->depth/2) + ((j+1)*depthInc) );
            glEnd();
        }
    }
    for (int i = 0; i < landforms.size(); i++){
        landforms[i]->draw(modeType, UP);
//        landforms[i]->draw(false, showSilhouette, showShadow);
//        
//        Trace t;        
//        t = landforms[i]->getBase();
//        glPointSize(5);
//        glColor3f(0.0, 0.0, 1.0);
//        glBegin(GL_POINTS);
//        for (int j = 0; j < t.size(); j++){
//            glVertex3f( t[j].getX(), t[j].getY()+0.1, t[j].getZ());
//        }
//        glEnd();

//        vector<Point> pts = landforms[i]->getMaxis();
//        glPointSize(5);
//        glColor3f(0.8, 0.2, 0.5);
//        glBegin(GL_POINTS);
//        for (int j = 0; j < pts.size(); j++){
//            glVertex3f( pts[j].getX(), pts[j].getY()+0.101, pts[j].getZ());
//        }
//        glEnd();

//        vector< pair<Point, Point> > edges;
//        edges =  landforms[i]->getBase().getSkeleton().getTopologicalEdgesPoints();
//        glLineWidth(5);
//        glColor3f(0.0, 1.0, 1.0);
//        for (int j = 0; j < edges.size(); j++){
//            Point p1 = edges[j].first;
//            Point p2 = edges[j].second;
//            glBegin(GL_LINE_STRIP);
//            glVertex3f( p1.getX(), p1.getY()+2.1, p1.getZ());
//            glVertex3f( p2.getX(), p2.getY()+2.1, p2.getZ());
//            glEnd();
//        }        
//        vector< pair<TopologicalVertex*, int> > tops;
//        tops = landforms[i]->getBase().getSkeleton().getTopologicalVertices();
//        glInitNames();
//        glPushName(0);
//        for (int j = 0; j < tops.size(); j++){
//            Point p = tops[j].first->getVertex()->getPosition();
//            p.setY(2.0);
//            Sphere s(p, 0.15);
//            glLoadName(j);
//            s.draw();
//        }
    }
    if (base.size() > 5){
        glLineWidth(3);
        glColor3f(1.0, 0.0, 1.0);
        glBegin(GL_LINE_STRIP);
        for (int j = 0; j < base.size(); j++){
            glVertex3f( base[j].getX(), base[j].getY(), base[j].getZ());
        }
        glEnd();
    }
    if (silhouette.size() > 5){
        glLineWidth(3);
        glColor3f(1.0, 0.0, 1.0);
        glBegin(GL_LINE_STRIP);
        for (int j = 0; j < silhouette.size(); j++){
            glVertex3f(silhouette[j].getX(), silhouette[j].getY(), silhouette[j].getZ());
        }
        glEnd();
    }
    if (showPlane){
        plane.draw(1.0, 1.0, 1.0);
    }
    if (showMesh) drawMesh();
    if (modeType == CONE){
        int size;
        int r,g,b;
        for(int i = 0; i < landforms.size(); i++){
            Base b = landforms[i]->getBase();
            size = b.size();
            printf("Land: %i   Size: %i\n",landforms.size(), size);
            for (int j = 0; j < size; j++){
             Cone c(b[j], 1000, Point(0.0,1000.0,0.0));
             glColor3f(float(j)/float(size), float(j)/float(size+1), float(j)/float(size+2));
             c.draw();
            }
        }
    }
}

void Terrain::drawMesh(){
    for (FaceIter f = mesh.faces_begin(); f != mesh.faces_end(); f++) {
        Pvert *v0 = Pvert::cast((*f)->vertex(0));
        Pvert *v1 = Pvert::cast((*f)->vertex(1));
        Pvert *v2 = Pvert::cast((*f)->vertex(2));
//        Point normal = (v0->position - v1->position).crossProduct(v0->position - v2->position);
//        if (-camera.dotProduct(normal) <= 0){
            glLineWidth(1);

            if(modeType == SHADING)
                glBegin(GL_TRIANGLES);
            else
                glBegin(GL_LINE_LOOP);
            Point n = v0->getNormal().normalize();

            glColor3ub(235, 180, 130);
            if (colorize) 
                glColor3ub(int(v0->getColor().getX()), int(v0->getColor().getY()), int(v0->getColor().getZ()));
            glNormal3f(-n.getX(), -n.getY(), -n.getZ());
            v0->position.draw();

            if (colorize) 
                glColor3ub(int(v1->getColor().getX()), int(v1->getColor().getY()), int(v1->getColor().getZ()));
            n = v1->getNormal().normalize();
            glNormal3f(-n.getX(), -n.getY(), -n.getZ());
            v1->position.draw();

            if (colorize) 
                glColor3ub(int(v2->getColor().getX()), int(v2->getColor().getY()), int(v2->getColor().getZ()));
            n = v2->getNormal().normalize();
            glNormal3f(-n.getX(), -n.getY(), -n.getZ());
            v2->position.draw();
            glEnd();
//        }
    }
}
