#include "canvas.h"

//////////////
//  Canvas  //
//////////////

Canvas::Canvas(QWidget* parent) : QGLWidget(parent){
    this->interval = 50;
    this->idTimer = this->startTimer(this->interval);
    scene = NULL;
}

Canvas::~Canvas(){

}

void Canvas::timerEvent(QTimerEvent *event){
    glDraw();
}

void Canvas::setRotation(float x, float y, float z){
    this->angle_x = x;
    this->angle_y = y;
    this->angle_z = z;
    if (this->scene) this->scene->setRotation(angle_x, angle_y, angle_z);
}

void Canvas::setTranslation(float x, float y, float z){
    translation.setX(x);
    translation.setY(y);
    translation.setZ(z);
    if (this->scene) this->scene->setTranslation(x, y, z);
}

void Canvas::setCamera( float posX, float posY, float posZ, 
                float latX, float latY, float latZ, 
                float upX,  float upY,  float upZ,
                float fov, float aspect, bool perspective)
{
    this->camera.setPosition(posX, posY, posZ);
    this->camera.setLookat(latX, latY, latZ);
    this->camera.setUp(upX, upY, upZ);
    this->camera.setFov(fov);
    this->camera.setAspectRatio(aspect);
    
    if (perspective) this->camera.enablePerspective();
    else this->camera.disablePerspective();
}

Camera Canvas::getCamera(){
    return camera;
}

void Canvas::setWindow(float l, float r, float b, float t, float n, float f){
    this->camera.setWindow(l, r, b, t, n, f);
}

void Canvas::setViewport(int _left, int _top, int _width, int _height){
    this->camera.setViewport(_left, _top, _width, _height);
}

void Canvas::enableFocus(){
    setFocusPolicy(Qt::WheelFocus);
}

void Canvas::setScene(Scene* s){
    scene = s;
}

Scene* Canvas::getScene(){
    return scene;
}

void Canvas::resizeGL(int _width, int _height){
    camera.resize(_width, _height);
}

void Canvas::setTitle(string _title){
    title = _title;
}

///////////////////
//  SceneCanvas  //
///////////////////

SceneCanvas::SceneCanvas(QWidget* parent) : Canvas(parent){}
SceneCanvas::~SceneCanvas(){}


void SceneCanvas::initializeGL(){
    GLfloat  dimLight[] = { 0.3f, 0.3f, 0.3f, 1.0f };
    GLfloat  sourceLight[] = { 0.95f, 0.95f, 0.95f, 10.0f };
    GLfloat	 lightPos[] = { 0.0f, 50.0f, 0.0f, 10.0f };

    glEnable(GL_DEPTH_TEST);

    glEnable(GL_LIGHTING);

    glLightfv(GL_LIGHT0, GL_AMBIENT, dimLight);
    glLightfv(GL_LIGHT0,GL_DIFFUSE,sourceLight);
    glLightfv(GL_LIGHT0,GL_POSITION,lightPos);
    glEnable(GL_LIGHT0);

    glEnable(GL_COLOR_MATERIAL);

    glClearColor(0.80f, 0.80f, 0.80f, 1.0f );
}

///////////////////
//  ImageCanvas  //
///////////////////

ImageCanvas::ImageCanvas(QWidget* _par) : Canvas(_par){

}

ImageCanvas::~ImageCanvas(){

}

void ImageCanvas::initializeGL(){
    glClearColor(0.10f, 0.10f, 0.10f, 1.0f );
}

void ImageCanvas::paintGL(){
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    updateImage();
    QPainter painter(this);
    painter.drawImage(QRect(0,0,camera.viewport.width,camera.viewport.height), this->image);

    glColor3f(1.0, 1.0, 0.0);
    renderText(5, 15, QString(title.data()));

    glFlush();
}

////////////////////
//  SketchCanvas  //
////////////////////

SketchCanvas::SketchCanvas(QWidget* _par) : Canvas(_par){

}

SketchCanvas::~SketchCanvas(){

}

void SketchCanvas::initializeGL(){
    glClearColor(0.80f, 0.80f, 0.80f, 1.0f );
}

void SketchCanvas::paintGL(){
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    updateSketch();

    glBegin(GL_LINE_STRIP);
    glColor3f(1.0, 0.0, 0.0);
    glLineWidth(5);

    for(int i = 0; i < trace.size(); i++){
        glVertex3f(trace[i].getX(), trace[i].getY(), trace[i].getZ());
    }
    glEnd();
    
    glColor3f(0.0, 0.0, 0.0);
    renderText(5, 15, QString(title.data()));
    
    glFlush();
}


//////////////////
//  DrawCanvas  //
//////////////////

DrawCanvas::DrawCanvas(QWidget* _par) : ImageCanvas(_par){

}

DrawCanvas::~DrawCanvas(){

}

QImage DrawCanvas::saveImage(){
    int gray;
    
    glDraw();
    QImage image = grabFrameBuffer();
    QImage newImage(100,100, QImage::Format_RGB16);
    for (int i = 0; i < 100; i++){
        for (int j = 0; j < 100; j++){
            gray = image.pixel(i, 99-j);
            newImage.setPixel(i, j, qRgb(gray, gray, gray));
        }
    }
    return newImage;
}

void DrawCanvas::initializeGL(){
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f );
}

void DrawCanvas::paintGL(){
    if (points.size() > 0){
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glClearColor(1.0f, 1.0f, 1.0f, 1.0f );
        glColor3f(0.0f, 0.0f, 0.0f); 
        
        GLUtesselator *pTess;
        pTess = gluNewTess();

        gluTessCallback(pTess, GLU_TESS_VERTEX, (GLvoid(*)()) &glVertex3dv);
        gluTessCallback(pTess, GLU_TESS_BEGIN, (GLvoid(*)()) &glBegin);
        gluTessCallback(pTess, GLU_TESS_END, (GLvoid(*)()) &glEnd);
        
        gluTessBeginPolygon(pTess, NULL);
        gluTessBeginContour(pTess);

        GLdouble p[this->points.size()][3];
        for(int i = 0; i < this->points.size(); i++){
            p[i][0] = (points[i].getX() + 12.5)*100.0/640.0;
            p[i][1] = (points[i].getZ() + 12.5)*100.0/480.0;
            p[i][2] = 0.0;
            gluTessVertex(pTess, p[i], p[i]); 
        }
            
        gluTessEndContour(pTess);
        gluTessEndPolygon(pTess);
        
        gluDeleteTess(pTess);
        glFlush();
    }
    glColor3f(1.0, 1.0, 1.0);
    renderText(5, 15, QString(title.data()));
}
