#include "interpreter.h"
#include "mysuggestion.h"
#include "myannotation.h"
#include "myscene.h"

///////////////////
//  Interpreter  //
///////////////////

Interpreter::Interpreter(MySuggestion* sug, Terrain& _terrain) : terrain(_terrain){
    this->suggestion = sug;
    setWindow();
}

Interpreter::~Interpreter(){

}

void Interpreter::executeAction(int i){
    actions[i]->act();
}

void Interpreter::drawInvisiblePlane(Point first, Point last){
    Point v1 = first + (last - first) * 200;
    Point v2 = last + (first - last) * 200;
    
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glBegin(GL_POLYGON);
        glVertex3f(v1.getX(), v1.getY(),v1.getZ());
        glVertex3f(v1.getX(), v1.getY()+400,v1.getZ());
        glVertex3f(v2.getX(), v2.getY()+400,v2.getZ());
        glVertex3f(v2.getX(), v2.getY(),v2.getZ());
    glEnd();
}

void Interpreter::addAction(int key, Action* action){
    this->actions[key] = action;
}

Point Interpreter::projectionPoint(Point point){
    GLdouble x, y, z;
    int viewportMatrix[4];
    double modelview[16];
    double projection[16];
    float p[1];    
    
    glGetIntegerv(GL_VIEWPORT, viewportMatrix);
    glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
    glGetDoublev(GL_PROJECTION_MATRIX, projection);
    
    glReadPixels(int(point.getX()), this->viewHeight - int(point.getY()), 
                 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &p);
    
    gluUnProject(int(point.getX()), this->viewHeight - int(point.getY()), p[0], 
                 modelview, projection, viewportMatrix, &x, &y, &z);
                 
    return Point(x,y+0.1,z);
}


void Interpreter::setViewport(int _width, int _height){
    this->viewWidth = _width;
    this->viewHeight = _height;
}

void Interpreter::setWindow(){
    this->windLeft = -5;
    this->windRight = 5;
    this->windBottom = -5;
    this->windTop = 5;
}

/////////////////////
//  UpInterpreter  //
/////////////////////

UpInterpreter::UpInterpreter(MySuggestion* sug, Terrain& ter) : Interpreter(sug, ter){
    annotation = NULL;
    terrain.startBase();
}

UpInterpreter::~UpInterpreter(){

}

void UpInterpreter::setAnnotation(MyAnnotation* note){
    annotation = note;
}

void UpInterpreter::startTrace(){
    cursors.clear();

}

void UpInterpreter::addCursor(Point c){
    if (cursors.size() == 0) terrain.startBase();
    Point scenePoint = projectionPoint(c);
    scenePoint.setY(0.01);
    cursors.push_back(scenePoint);
    terrain.addBasePoint(scenePoint);
}

void UpInterpreter::finishTrace(){
    if (cursors.size() > 5){
        terrain.finishBase();
    
//        Landform* land = terrain.createLandform(STEP_PLATEU);
//        Landform* land = terrain.createLandform(RAMP_PLATEU);
//        Landform* land = terrain.createLandform(SMOOTH_PLATEU);
//        Landform* land = terrain.createLandform(MAXISDISTANCE_HILL);
//        Landform* land = terrain.createLandform(DISTANCE_HILL);
//        Landform* land = terrain.createLandform(NORMALIZED_HILL);
//        Landform* land = terrain.createLandform(VOLCANO);
//        Landform* land = terrain.createLandform(ISLAND);
//        land->setHeight(2.0);
//
//        Trace trace = terrain.getBase();
//        if ( trace.size() > 3 ){
//            for (int i = 0; i < trace.size(); i++){
//                land->addBasePoint(trace[i]);
//            }
//            terrain.updateTerrain();
//        }

        startTrace();

        if (showHill) this->suggestion->generateSuggestion(DISTANCE_HILL_GEN);
        if (showPlateu) this->suggestion->generateSuggestion(SMOOTH_PLATEU_GEN);
        if (showPlateu) this->suggestion->generateSuggestion(MAXIS_PLATEU_GEN);
        if (showVolcano) this->suggestion->generateSuggestion(VOLCANO_GEN);
        if (showIsland) this->suggestion->generateSuggestion(ISLAND_GEN);
//
//        this->suggestion->generateSuggestion(MAXISDISTANCE_HILL_GEN);
//        this->suggestion->generateSuggestion(NORMALIZED_HILL_GEN);
//
//        this->suggestion->generateSuggestion(RAMP_PLATEU_GEN);
//        this->suggestion->generateSuggestion(STEP_PLATEU_GEN);
    }
    else{
        terrain.startBase();
        startTrace();
    }

}


////////////////////////
//  FrontInterpreter  //
////////////////////////

FrontInterpreter::FrontInterpreter(MySuggestion* sug, Terrain& ter) : Interpreter(sug, ter){
    terrain.startSilhouette();
    moving = false;
}

FrontInterpreter::~FrontInterpreter(){
    
}

bool FrontInterpreter::processPick(Point cursor){
    printf("Processing\n");
    vector<Selected> sel = canvas->getScene()->selectElements(cursors[0].getX(), cursors[0].getY());
    bool picked= false;
    unsigned long distance = 4294967294;  
    
    for (int i = 0; i < sel.size(); i++){
        if (sel[i].id >= 20 and sel[i].id <= 100 and distance > sel[i].zmin){
            selected = sel[i];
            distance = sel[i].zmin;
            picked = true;
        }
    }
    return picked;
}


void FrontInterpreter::moveTopologicalPoint(){
    selected.show();
    Point p = canvas->getScene()->getScenePoint(cursors[0]);
    p.show();
    p.setY(0.0);
    
    p.show();
    Landform* land = terrain.getLandform(p);
    if (land != NULL){
        printf("Land: %f\n\n", land->getHeight());
        
        Skeleton skel = land->getBase().getSkeleton();

        TopologicalSkeleton& topSkel = skel.getTopologicalSkeleton();
        top = topSkel.getVertex(selected.id - 20);

        skel.updateGeometricHeights();
    }
}

void FrontInterpreter::startTrace(){
    cursors.clear();

    terrain.startSilhouette();
}

void FrontInterpreter::addCursor(Point c){
    cursors.push_back(c);
    if (cursors.size() == 1 and processPick(c)){    
        MyScene* s = (MyScene*) canvas->getScene();
        s->enablePlane();
        s->setPlane(Point(0.0,0.0,1.0), projectionPoint(c));
        moving = true;
        
        Point p = canvas->getScene()->getScenePoint(cursors[0]);
        p.show();
        p.setY(0.0);
        
        Landform* land = terrain.getLandform(p);
        if (land != NULL){
            printf("Land: %f\n\n", land->getHeight());
            
            vector< Point > skelPoints = land->getBase().getSkeleton().getTopologicalPoints();
            for (int i = 0; i < skelPoints.size(); i++){
                printf("%i   ",i);skelPoints[i].show();
            }
            
            
            TopologicalSkeleton& topSkel = land->getBase().getSkeleton().getTopologicalSkeleton();
            top = topSkel.getVertex(selected.id - 20);
            printf("\nSelected: %i\n", selected.id - 20);selected.show();
            
            oldVertex = skelPoints[selected.id - 20];
            printf("Old:  ");oldVertex.show();
            
//            oldVertex = top->getVertex()->getPosition();
        }
        else{
            moving = false;
            s->disablePlane();
            printf("\nin Add Cursor\nClick mal sucedido. Tente novamente!!!\n\n");
        }
    }
    else if (moving){
        top->setHeight(projectionPoint(c).getY());
//        printf("Top height: %f\n",top->getHeight());
    }
}

void FrontInterpreter::finishTrace(){
    printf("Finish\n");
    if ( processPick(cursors[0]) or moving){
        moveTopologicalPoint();
        
        if (moving) terrain.updateTerrain();
        
        moving = false;
        
        MyScene* s = (MyScene*) canvas->getScene();
        s->disablePlane();
    }
    else if (cursors.size() >= 5 ){
        printf("\n\nSilhouette\n\n");
        Point first = projectionPoint(cursors[0]);
        Point last  = projectionPoint(cursors.back());
        drawInvisiblePlane(first, last);
        
        Point scenePoint; Point lastAdded = first;;
        for (int i = 0; i < cursors.size(); i++){
            scenePoint = projectionPoint(cursors[i]);
            if ((scenePoint - lastAdded).lenght() < 3.0){
                scenePoint.setId(SILHOUETTE);
                terrain.addSilhouettePoint(scenePoint);
                lastAdded = scenePoint;
            }
        }
        terrain.adjustSilhouette();
        
        executeAction(HAROLD_GEN);

        this->suggestion->generateSuggestion(UPDATE_SILHOUETTE);
    }
}

